#pragma once

#include "TdZdd/include/tdzdd/DdSpec.hpp"
#include "TdZdd/include/tdzdd/DdStructure.hpp"
#include <algorithm>
#include <bitset>
#include <iostream>
#include <map>
#include <random>
#include <time.h>
#include <utility>
#include <vector>

#include "input.cpp"
#include "warshall_floyd.cpp"

// v, u を from, to

class AllPair {
  public:
    class SimPath : public tdzdd::PodArrayDdSpec<SimPath, short, 2> {
        std::vector<Edge> edges;
        int m;
        int n;
        int s, t;
        int l;
        std::vector<std::vector<int>> counter;
        std::vector<std::vector<int>> dist;

        int c_level;

      public:
        SimPath(int n, int m, std::vector<Edge> edges, int l, std::vector<std::vector<int>> dist)
            : n(n), m(m), edges(edges), s(s), t(t), l(l), dist(dist) {
            setArraySize(n + 1);
            counter = std::vector<std::vector<int>>(m, std::vector<int>(n));
            for(int i = 0; i < m; i++) {
                counter[0][edges[i].from]++;
                counter[0][edges[i].to]++;
            }
            for(int i = 0; i < m - 1; i++) {
                counter[i][edges[i].from]--;
                counter[i][edges[i].to]--;

                counter[i + 1] = counter[i];
            }
            counter.back()[edges.back().from]--;
            counter.back()[edges.back().to]--;
        }

        int getRoot(short *mate) {
            for(int i = 0; i < n; i++) {
                mate[i] = -1;
            }
            mate[n] = 0;
            c_level = m + 1;
            return m;
        }

        int getChild(short *mate, int level, int value) {
            if(value == 1 && mate[n] >= l) {
                return 0;
            }

            Edge e = edges[m - level];

            auto is_independ = [&](int v) { // 独立点かどうか判定する関数
                return mate[v] == -1 || mate[v] == v + 1;
            };

            auto is_terminated = [&](int v) { // 端点かっどうか判定する関数
                return mate[v] > 0 && mate[v] != v + 1;
            };

            auto is_in_path = [&](int v) { // パスの中間点かどうか判定する関数
                return mate[v] == 0;
            };

            auto is_out_from_frontier = [&](int v) {
                return mate[v] == -1;
            };

            if(is_independ(e.from) && is_independ(e.to)) { // 両端点が独立点または未処理
                if(value == 1) {
                    mate[e.from] = e.to + 1;
                    mate[e.to] = e.from + 1;
                } else {
                    mate[e.from] = e.from + 1;
                    mate[e.to] = e.to + 1;
                }
            } else if(is_in_path(e.to)) { // e.to が中間点
                if(value == 1) {
                    return 0;
                } else {
                    if(mate[e.from] == -1)
                        mate[e.from] = e.from + 1;
                }
            } else if(is_in_path(e.from)) { // e.from が中間点
                if(value == 1) {
                    return 0;
                } else {
                    if(mate[e.to] == -1)
                        mate[e.to] = e.to + 1;
                }
            } else if(is_independ(e.from) && is_terminated(e.to)) { // e.from が独立点または未処理かつ、e.to が端点
                if(value == 1) {
                    if(!is_out_from_frontier(mate[e.to] - 1)) {
                        mate[mate[e.to] - 1] = e.from + 1;
                    }
                    mate[e.from] = mate[e.to];
                    mate[e.to] = 0;
                } else {
                    mate[e.from] = e.from + 1;
                }
            } else if(is_terminated(e.from) && is_independ(e.to)) { // e.to が独立点または未処理かつ、e.from が端点
                if(value == 1) {
                    if(!is_out_from_frontier(mate[e.from] - 1)) {
                        mate[mate[e.from] - 1] = e.to + 1;
                    }
                    mate[e.to] = mate[e.from];
                    mate[e.from] = 0;
                } else {
                    mate[e.to] = e.to + 1;
                }
            } else if(mate[e.to] == e.from + 1 && mate[e.from] == e.to + 1) { // パスの端点を繋いでしまうようなやつ
                if(value == 1) {
                    return 0;
                }
            } else if(is_terminated(e.from) && is_terminated(e.to)) { // パスとパスをつなぐようなケース
                if(value == 1) {
                    if(is_out_from_frontier(mate[e.from] - 1) && is_out_from_frontier(mate[e.to] - 1)) {
                        bool is_path = true;
                        for(int i = 0; i < n; i++) {
                            if(i != e.from && i != e.to) {
                                if(is_terminated(i)) {
                                    is_path = false;
                                    break;
                                }
                            }
                        }

                        if(is_path) {
                            return -1;
                        } else {
                            return 0;
                        }
                    }
                    if(!is_out_from_frontier(mate[e.from] - 1)) {
                        mate[mate[e.from] - 1] = mate[e.to];
                    }
                    if(!is_out_from_frontier(mate[e.to] - 1)) {
                        mate[mate[e.to] - 1] = mate[e.from];
                    }

                    mate[e.from] = 0;
                    mate[e.to] = 0;
                }
            }
            std::vector<std::pair<int, int>> pathes;
            bool is_calc = false;
            int terminal_cnt = 0;

            auto generate_path_pairs = [&]() {
                for(int i = 0; i < n; i++) {
                    if(mate[i] > 0 && mate[i] - 1 != i) {
                        int x = mate[i] - 1;
                        int y = i;

                        if(x > y)
                            std::swap(x, y);

                        pathes.push_back(std::make_pair(x, y));
                    }
                }

                std::sort(pathes.begin(), pathes.end());
                pathes.erase(std::unique(pathes.begin(), pathes.end()), pathes.end());

                for(int i = 0; i < pathes.size(); i++) {
                    if(mate[pathes[i].first] < 0 || mate[pathes[i].second] < 0) {
                        terminal_cnt++;
                    }
                }
                is_calc = true;
            };

            if(counter[m - level][e.from] == 0) {
                if(!is_calc) {
                    generate_path_pairs();
                }

                if(is_terminated(e.from) && terminal_cnt <= 1) {
                    terminal_cnt++;
                } else if(is_terminated(e.from)) {
                    return 0;
                }

                if(terminal_cnt == 2 && is_terminated(e.from)) {
                    if(is_out_from_frontier(mate[e.from] - 1)) { // パスの片方が既にフロンティアの外
                        bool is_path = true;

                        for(int i = 0; i < n; i++) {
                            if(e.from != i) {
                                if(is_terminated(i)) {
                                    is_path = false;
                                    break;
                                }
                            }
                        }

                        if(is_path) {
                            return -1;
                        } else {
                            return 0;
                        }
                    }
                }

                if(terminal_cnt >= 3) {
                    return 0;
                }

                mate[e.from] = -1;
            }

            if(counter[m - level][e.to] == 0) {
                if(!is_calc) {
                    generate_path_pairs();
                }

                if(is_terminated(e.to) && terminal_cnt <= 1) {
                    terminal_cnt++;
                } else if(is_terminated(e.to)) {
                    return 0;
                }

                if(terminal_cnt == 2 && is_terminated(e.to)) {
                    if(is_out_from_frontier(mate[e.to] - 1)) { // パスの片方が既にフロンティアの外
                        bool is_path = true;

                        for(int i = 0; i < n; i++) {
                            if(e.to != i) {
                                if(is_terminated(i)) {
                                    is_path = false;
                                    break;
                                }
                            }
                        }

                        if(is_path) {
                            return -1;
                        } else {
                            return 0;
                        }
                    }
                }

                if(terminal_cnt >= 3) {
                    return 0;
                }

                mate[e.to] = -1;
            }

            if(value == 1) {
                mate[n]++;
            }

            if(l - mate[n] > 0) {
                if(!is_calc) {
                    generate_path_pairs();
                }

                std::vector<std::pair<int, int>> fillterd_pathes, terminal_pathes;
                for(int i = 0; i < pathes.size(); i++) {
                    if(mate[pathes[i].first] > 0 && mate[pathes[i].second] > 0) {
                        fillterd_pathes.push_back(pathes[i]);
                    } else {
                        terminal_pathes.push_back(pathes[i]);
                    }
                }

                int s = -1, t = -1;
                for(int i = 0; i < terminal_pathes.size(); i++) {
                    if(s == -1) {
                        if(mate[terminal_pathes[i].first] < 0) {
                            s = terminal_pathes[i].second;
                        } else {
                            s = terminal_pathes[i].first;
                        }
                    } else {
                        if(mate[terminal_pathes[i].first] < 0) {
                            t = terminal_pathes[i].second;
                        } else {
                            t = terminal_pathes[i].first;
                        }
                    }
                }

                if(fillterd_pathes.size() != 0) {
                    if(fillterd_pathes.size() <= 10) {
                        int mi = 1e9;
                        do {
                            int need_edges = 0;

                            if(s != -1) {
                                need_edges += std::min(
                                    dist[s][fillterd_pathes.front().first],
                                    dist[s][fillterd_pathes.front().second]);
                            }
                            for(int i = 0; i < (int)fillterd_pathes.size() - 1; i++) {
                                need_edges += std::min(
                                    {
                                        dist[fillterd_pathes[i].first][fillterd_pathes[i + 1].first],
                                        dist[fillterd_pathes[i].first][fillterd_pathes[i + 1].second],
                                        dist[fillterd_pathes[i].second][fillterd_pathes[i + 1].first],
                                        dist[fillterd_pathes[i].second][fillterd_pathes[i + 1].second],
                                    });
                            }

                            if(t != -1) {
                                need_edges += std::min(
                                    dist[t][fillterd_pathes.back().first],
                                    dist[t][fillterd_pathes.back().second]);
                            }

                            mi = std::min(mi, need_edges);
                        } while(std::next_permutation(fillterd_pathes.begin(), fillterd_pathes.end()));

                        int remain = l - mate[n];

                        if(mi > remain) {
                            return 0;
                        }
                    }
                }
            }

            return level - 1;
        }
    };
    void opt_edges_ordering(std::vector<Edge> &edges, int n) {
        struct state {
            int prev;
            int layer;
            long long score;
            int edge_idx;
            std::vector<long long> bits;
            std::vector<long long> frontier;

            bool is_expand;

            void resize_bitset(int m) {
                bits.resize((m + 64 - 1) / 64);
            }
            void set_bit(int i, bool b) {
                if(b) {
                    bits[i / 64] |= (1LL << (i % 64));
                } else {
                    bits[i / 64] &= ~(1LL << (i % 64));
                }
            }
            bool get_bit(int i) {
                return bits[i / 64] & (1LL << (i % 64));
            }

            void resize_frontier(int n) {
                frontier.resize((n + 64 - 1) / 64);
            }
            void set_frontier(int i, bool b) {
                if(b) {
                    frontier[i / 64] |= (1LL << (i % 64));
                } else {
                    frontier[i / 64] &= ~(1LL << (i % 64));
                }
            }
            bool get_frontier(int i) {
                return frontier[i / 64] & (1LL << (i % 64));
            }

            state(int m, int n) {
                // bits.resize(m, 0);
                resize_bitset(m);
                resize_frontier(n);
                is_expand = false;
                prev = -1;
                score = 0;
            }
            state() {
                is_expand = false;
                prev = -1;
                score = 0;
            }

            state &operator=(const state &a) {
                prev = a.prev;
                layer = a.layer;
                score = a.score;
                bits = a.bits;
                frontier = a.frontier;
                edge_idx = a.edge_idx;
                is_expand = a.is_expand;

                return *this;
            }
        };

        std::vector<std::vector<state>> states(edges.size());

        std::vector<std::vector<Edge>> edges_list(n);
        state initial_state(edges.size(), n);

        std::vector<int> deg(n);

        for(auto e : edges) {
            edges_list[e.from].push_back(e);
            edges_list[e.to].push_back(e);
            deg[e.from]++;
            deg[e.to]++;
        }

        int idx = -1;
        {
            int ma_deg = 1e9;
            for(int i = 0; i < n; i++) {
                if(deg[i] < ma_deg) {
                    ma_deg = deg[i];
                    idx = i;
                }
            }
        }

        int s = idx;

        for(int i = 0; i < edges_list[s].size(); i++) {
            state n_state = initial_state;
            n_state.set_bit(edges_list[s][i].idx, 1);
            n_state.layer = 0;
            n_state.score = 0;

            n_state.set_frontier(edges_list[s][i].from, 1);
            n_state.set_frontier(edges_list[s][i].to, 1);

            bool is_finished = true;
            for(int j = 0; j < edges_list[edges_list[s][i].from].size(); j++) {
                if(n_state.get_bit(edges_list[edges_list[s][i].from][j].idx) == 0) {
                    is_finished = false;
                }
            }

            if(is_finished)
                n_state.set_frontier(edges_list[s][i].from, 0);

            is_finished = true;
            for(int j = 0; j < edges_list[edges_list[s][i].to].size(); j++) {
                if(n_state.get_bit(edges_list[edges_list[s][i].to][j].idx) == 0) {
                    is_finished = false;
                }
            }

            if(is_finished)
                n_state.set_frontier(edges_list[s][i].to, 0);

            int frontier_count = 0;

            if(n_state.get_frontier(edges_list[s][i].from) == 1)
                frontier_count++;

            if(n_state.get_frontier(edges_list[s][i].to) == 1)
                frontier_count++;

            n_state.score += frontier_count * frontier_count * frontier_count;
            n_state.edge_idx = edges_list[s][i].idx;

            states[0].push_back(n_state);
        }

        int ChokudaiWidth = 1;

        clock_t start = clock();

        int timelimit = 100;

        std::map<std::vector<long long>, int> idx_mp;

        while(double(clock() - start) / CLOCKS_PER_SEC <= timelimit) {
            int state_sum = 0;
            for(int i = 0; i < (int)edges.size() - 1; i++) {
                if(states[i].size() == 0)
                    break;
                for(int j = 0; j < ChokudaiWidth; j++) {
                    int score = 1e9;
                    int idx = -1;

                    for(int k = 0; k < states[i].size(); k++) {
                        if(states[i][k].score <= score && states[i][k].is_expand == false) {
                            score = states[i][k].score;
                            idx = k;
                        }
                    }

                    if(idx == -1) {
                        break;
                    }

                    for(int k = 0; k < n; k++) {
                        if(states[i][idx].get_frontier(k) == 1) {
                            for(Edge e : edges_list[k]) {
                                if(states[i][idx].get_bit(e.idx) == 0) {
                                    state n_state = states[i][idx];
                                    n_state.set_bit(e.idx, 1);

                                    int eq_itr = -1;

                                    if(idx_mp.find(n_state.bits) != idx_mp.end()) {
                                        eq_itr = idx_mp[n_state.bits];
                                    }

                                    n_state.layer = states[i][idx].layer + 1;
                                    n_state.set_frontier(e.from, 1);
                                    n_state.set_frontier(e.to, 1);

                                    bool is_finished = true;
                                    for(int i = 0; i < edges_list[e.from].size(); i++) {
                                        if(n_state.get_bit(edges_list[e.from][i].idx) == 0) {
                                            is_finished = false;
                                            break;
                                        }
                                    }

                                    if(is_finished)
                                        n_state.set_frontier(e.from, 0);

                                    is_finished = true;
                                    for(int i = 0; i < edges_list[e.to].size(); i++) {
                                        if(n_state.get_bit(edges_list[e.to][i].idx) == 0) {
                                            is_finished = false;
                                            break;
                                        }
                                    }

                                    if(is_finished)
                                        n_state.set_frontier(e.to, 0);

                                    int frontier_size = 0;
                                    for(int l = 0; l < n_state.frontier.size(); l++) {
                                        frontier_size += __builtin_popcountll(n_state.frontier[l]);
                                    }

                                    n_state.score += frontier_size * frontier_size * frontier_size;

                                    n_state.prev = idx;

                                    n_state.edge_idx = e.idx;

                                    if(eq_itr != -1) {
                                        if(n_state.score < states[i + 1][eq_itr].score) {
                                            states[i + 1][eq_itr] = n_state;
                                            states[i + 1][eq_itr].is_expand = false;
                                        }
                                    } else {
                                        idx_mp[n_state.bits] = states[i + 1].size();

                                        states[i + 1].push_back(n_state);
                                        states[i + 1].back().is_expand = false;
                                    }
                                }
                            }
                        }
                    }

                    states[i][idx].is_expand = true;
                    state_sum += states[i].size();
                }
            }
        }
        int current_idx = 0;

        std::vector<int> order;

        int ma_frontier = 0;
        for(int layer = edges.size() - 1; layer >= 0; layer--) {
            order.push_back(states[layer][current_idx].edge_idx);

            if(layer != 0) {
                ma_frontier = std::max(int(pow(states[layer][current_idx].score - states[layer - 1][states[layer][current_idx].prev].score, (double)1 / 3)), ma_frontier);
            }
            current_idx = states[layer][current_idx].prev;
        }

        std::reverse(order.begin(), order.end());

        std::vector<Edge> n_edges;

        for(int i = 0; i < order.size(); i++) {
            n_edges.push_back(edges[order[i]]);
        }

        edges = n_edges;
    }

    void solve(Problem problem) {
        std::vector<Edge> edges;
        int edge_idx = 0;
        for(int i = 0; i < problem.n; i++) {
            for(int j = 0; j < problem.edges[i].size(); j++) {
                if(i < problem.edges[i][j]) {
                    edges.push_back(Edge(i, problem.edges[i][j]));
                    edges.back().value = 1;
                    edges.back().idx = edge_idx;
                    edge_idx++;
                }
            }
        }

        std::vector<std::vector<int>> dist = warshall_floyd(edges, problem.n);

        opt_edges_ordering(edges, problem.n);

        SimPath simpath(problem.n, problem.m, edges, problem.l, dist);
        tdzdd::DdStructure<2> dd(simpath, true);

        std::cout << dd.zddCardinality() << std::endl;
    }
};
